"""
feature_extractor.py

Helper class to encapsulate feature extraction methods

"""
import pandas as pd
from nkululeko.util import Util 
from nkululeko.feat_extract.feats_opensmile import Opensmileset


class FeatureExtractor:
    """
    Extract acoustic features from audio samples, using several feature extractors (appends the features column-wise)
    Args:
        data_df (pandas.DataFrame): dataframe with audiofile paths as index
        feats_types (array of strings): designations of acoustic feature extractors to be used
        data_name (string): names of databases that are extracted (for the caching)
        feats_designation (string): the type of split (train/test), also is used for the cache name.    
    Returns:
        df (pandas.DataFrame): dataframe with same index as data_df and acoustic features in columns  
    """
    df = None # pandas dataframe to store the features (and indexed with the data from the sets)
    data_df = None # dataframe to get audio paths

# def __init__
    def __init__(self, data_df, feats_types, data_name, feats_designation):
        self.data_df = data_df
        self.data_name = data_name
        self.feats_types = feats_types
        self.util = Util('feature_extractor')
        self.feats_designation = feats_designation
    
    def extract(self):
        # feats_types = self.util.config_val_list('FEATS', 'type', ['os'])
        self.featExtractor = None
        self.feats= pd.DataFrame()
        _scale = True
        for feats_type in self.feats_types:
            store_name = f'{self.data_name}_{feats_type}'   
            if feats_type=='os':
                self.featExtractor = Opensmileset(f'{store_name}_{self.feats_designation}', self.data_df)
            elif feats_type=='trill':
                from nkululeko.feat_extract.feats_trill import TRILLset
                self.featExtractor = TRILLset(f'{store_name}_{self.feats_designation}', self.data_df)
            elif feats_type=='wav2vec':
                from nkululeko.feat_extract.feats_wav2vec2 import Wav2vec2
                self.featExtractor = Wav2vec2(f'{store_name}_{self.feats_designation}', self.data_df)
            elif feats_type=='audmodel':
                from nkululeko.feat_extract.feats_audmodel import AudModelSet
                self.featExtractor = AudModelSet(f'{store_name}_{self.feats_designation}', self.data_df)
            elif feats_type=='auddim':
                from nkululeko.feat_extract.feats_audmodel_dim import AudModelDimSet
                self.featExtractor = AudModelDimSet(f'{store_name}_{self.feats_designation}', self.data_df)
            elif feats_type=='agender':
                from nkululeko.feat_extract.feats_agender import AudModelAgenderSet
                self.featExtractor = AudModelAgenderSet(f'{store_name}_{self.feats_designation}', self.data_df)
            elif feats_type=='agender_agender':
                from nkululeko.feat_extract.feats_agender_agender import AgenderAgenderSet
                self.featExtractor = AgenderAgenderSet(f'{store_name}_{self.feats_designation}', self.data_df)
            elif feats_type=='snr':
                from nkululeko.feat_extract.feats_snr import SNRSet
                self.featExtractor = SNRSet(f'{store_name}_{self.feats_designation}', self.data_df)
            elif feats_type=='mos':
                from nkululeko.feat_extract.feats_mos import MOSSet
                self.featExtractor = MOSSet(f'{store_name}_{self.feats_designation}', self.data_df)
            elif feats_type=='squim':
                from nkululeko.feat_extract.feats_squim import SQUIMSet
                self.featExtractor = SQUIMSet(f'{store_name}_{self.feats_designation}', self.data_df)
            elif feats_type=='clap':
                from nkululeko.feat_extract.feats_clap import Clap
                self.featExtractor = Clap(f'{store_name}_{self.feats_designation}', self.data_df)
            elif feats_type=='praat':
                from nkululeko.feat_extract.feats_praat import Praatset
                self.featExtractor = Praatset(f'{store_name}_{self.feats_designation}', self.data_df)
            elif feats_type=='mld':
                from nkululeko.feat_extract.feats_mld import MLD_set
                self.featExtractor = MLD_set(f'{store_name}_{self.feats_designation}', self.data_df)
            elif feats_type=='import':
                from nkululeko.feat_extract.feats_import import Importset
                self.featExtractor = Importset(f'{store_name}_{self.feats_designation}', self.data_df)
            else:
                self.util.error(f'unknown feats_type: {feats_type}')

            self.featExtractor.extract()
            self.featExtractor.filter()
            # remove samples that were not extracted by MLD
            #self.df_test = self.df_test.loc[self.df_test.index.intersection(featExtractor_test.df.index)]
            #self.df_train = self.df_train.loc[self.df_train.index.intersection(featExtractor_train.df.index)]
            self.util.debug(f'{feats_type}: shape : {self.featExtractor.df.shape}')
            self.feats = pd.concat([self.feats, self.featExtractor.df], axis = 1)
        return self.feats

    def extract_sample(self, signal, sr):
        return self.featExtractor.extract_sample(signal, sr)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

###########################################################################
#                                                                         #
#  Praat Script Syllable Nuclei                                           #
#  Copyright (C) 2008  Nivja de Jong and Ton Wempe                        #
#                                                                         #
#    This program is free software: you can redistribute it and/or modify #
#    it under the terms of the GNU General Public License as published by #
#    the Free Software Foundation, either version 3 of the License, or    #
#    (at your option) any later version.                                  #
#                                                                         #
#    This program is distributed in the hope that it will be useful,      #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of       #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        #
#    GNU General Public License for more details.                         #
#                                                                         #
#    You should have received a copy of the GNU General Public License    #
#    along with this program.  If not, see http://www.gnu.org/licenses/   #
#                                                                         #
###########################################################################
#
# modified 2010.09.17 by Hugo Quené, Ingrid Persoon, & Nivja de Jong
# Overview of changes:
# + change threshold-calculator: rather than using median, use the almost maximum
#     minus 25dB. (25 dB is in line with the standard setting to detect silence
#     in the "To TextGrid (silences)" function.
#     Almost maximum (.99 quantile) is used rather than maximum to avoid using
#     irrelevant non-speech sound-bursts.
# + add silence-information to calculate articulation rate and ASD (average syllable
#     duration.
#     NB: speech rate = number of syllables / total time
#         articulation rate = number of syllables / phonation time
# + remove max number of syllable nuclei
# + refer to objects by unique identifier, not by name
# + keep track of all created intermediate objects, select these explicitly,
#     then Remove
# + provide summary output in Info window
# + do not save TextGrid-file but leave it in Object-window for inspection
#     (if requested in startup-form)
# + allow Sound to have starting time different from zero
#      for Sound objects created with Extract (preserve times)
# + programming of checking loop for mindip adjusted
#      in the orig version, precedingtime was not modified if the peak was rejected !!
#      var precedingtime and precedingint renamed to currenttime and currentint
#
# + bug fixed concerning summing total pause, feb 28th 2011
###########################################################################


# counts syllables of all sound utterances in a directory
# NB unstressed syllables are sometimes overlooked
# NB filter sounds that are quite noisy beforehand
# NB use Silence threshold (dB) = -25 (or -20?)
# NB use Minimum dip between peaks (dB) = between 2-4 (you can first try;
#                                                      For clean and filtered: 4)
#
#
# Translated to Python in 2019 by David Feinberg
# I changed all the variable names so they are human readable



import math
import pandas as pd
import parselmouth

from glob import glob
from parselmouth.praat import call


def speech_rate(filename):
    silencedb = -25
    mindip = 2
    minpause = 0.3
    sound = parselmouth.Sound(filename)
    originaldur = sound.get_total_duration()
    intensity = sound.to_intensity(50)
    start = call(intensity, "Get time from frame number", 1)
    nframes = call(intensity, "Get number of frames")
    end = call(intensity, "Get time from frame number", nframes)
    min_intensity = call(intensity, "Get minimum", 0, 0, "Parabolic")
    max_intensity = call(intensity, "Get maximum", 0, 0, "Parabolic")

    # get .99 quantile to get maximum (without influence of non-speech sound bursts)
    max_99_intensity = call(intensity, "Get quantile", 0, 0, 0.99)

    # estimate Intensity threshold
    threshold = max_99_intensity + silencedb
    threshold2 = max_intensity - max_99_intensity
    threshold3 = silencedb - threshold2
    if threshold < min_intensity:
        threshold = min_intensity

    # get pauses (silences) and speakingtime
    textgrid = call(intensity, "To TextGrid (silences)", threshold3, minpause, 0.1, "silent", "sounding")
    silencetier = call(textgrid, "Extract tier", 1)
    silencetable = call(silencetier, "Down to TableOfReal", "sounding")
    npauses = call(silencetable, "Get number of rows")
    speakingtot = 0
    for ipause in range(npauses):
        pause = ipause + 1
        beginsound = call(silencetable, "Get value", pause, 1)
        endsound = call(silencetable, "Get value", pause, 2)
        speakingdur = endsound - beginsound
        speakingtot += speakingdur

    intensity_matrix = call(intensity, "Down to Matrix")
    # sndintid = sound_from_intensity_matrix
    sound_from_intensity_matrix = call(intensity_matrix, "To Sound (slice)", 1)
    # use total duration, not end time, to find out duration of intdur (intensity_duration)
    # in order to allow nonzero starting times.
    intensity_duration = call(sound_from_intensity_matrix, "Get total duration")
    intensity_max = call(sound_from_intensity_matrix, "Get maximum", 0, 0, "Parabolic")
    point_process = call(sound_from_intensity_matrix, "To PointProcess (extrema)", "Left", "yes", "no", "Sinc70")
    # estimate peak positions (all peaks)
    numpeaks = call(point_process, "Get number of points")
    t = [call(point_process, "Get time from index", i + 1) for i in range(numpeaks)]

    # fill array with intensity values
    timepeaks = []
    peakcount = 0
    intensities = []
    for i in range(numpeaks):
        value = call(sound_from_intensity_matrix, "Get value at time", t[i], "Cubic")
        if value > threshold:
            peakcount += 1
            intensities.append(value)
            timepeaks.append(t[i])

    # fill array with valid peaks: only intensity values if preceding
    # dip in intensity is greater than mindip
    validpeakcount = 0
    currenttime = timepeaks[0]
    currentint = intensities[0]
    validtime = []

    for p in range(peakcount - 1):
        following = p + 1
        followingtime = timepeaks[p + 1]
        dip = call(intensity, "Get minimum", currenttime, timepeaks[p + 1], "None")
        diffint = abs(currentint - dip)
        if diffint > mindip:
            validpeakcount += 1
            validtime.append(timepeaks[p])
        currenttime = timepeaks[following]
        currentint = call(intensity, "Get value at time", timepeaks[following], "Cubic")

    # Look for only voiced parts
    pitch = sound.to_pitch_ac(0.02, 30, 4, False, 0.03, 0.25, 0.01, 0.35, 0.25, 450)
    voicedcount = 0
    voicedpeak = []

    for time in range(validpeakcount):
        querytime = validtime[time]
        whichinterval = call(textgrid, "Get interval at time", 1, querytime)
        whichlabel = call(textgrid, "Get label of interval", 1, whichinterval)
        value = pitch.get_value_at_time(querytime) 
        if not math.isnan(value):
            if whichlabel == "sounding":
                voicedcount += 1
                voicedpeak.append(validtime[time])

    # calculate time correction due to shift in time for Sound object versus
    # intensity object
    timecorrection = originaldur / intensity_duration

    # Insert voiced peaks in TextGrid
    call(textgrid, "Insert point tier", 1, "syllables")
    for i in range(len(voicedpeak)):
        position = (voicedpeak[i] * timecorrection)
        call(textgrid, "Insert point", 1, position, "")

    # return results
    speakingrate = voicedcount / originaldur
    articulationrate = voicedcount / speakingtot
    npause = npauses - 1
    try:
        asd = speakingtot / voicedcount
    except ZeroDivisionError:
        asd = 0
        print('caught zero division')
    speechrate_dictionary = {'file':file,
                             'nsyll':voicedcount,
                             'npause': npause,
                             'dur(s)':originaldur,
                             'phonationtime(s)':intensity_duration,
                             'speechrate(nsyll / dur)': speakingrate,
                             "articulation rate(nsyll / phonationtime)":articulationrate,
                             "ASD(speakingtime / nsyll)":asd}
    return speechrate_dictionary


def get_files():
    audio_dir = '/home/audeering.local/fburkhardt/audb/projectsmile-speakingrate-agent/3.0.0/fe182b91/audio/'
    files = glob(f'{audio_dir}/*.wav')
    files.extend(glob('test_voices/*.mp3'))
    files.extend(glob('test_voices/*.ogg'))
    files.extend(glob('test_voices/*.aiff'))
    files.extend(glob('test_voices/*.aifc'))
    files.extend(glob('test_voices/*.au'))
    files.extend(glob('test_voices/*.nist'))
    files.extend(glob('test_voices/*.flac'))
    return files


if __name__ == "__main__":
    files = get_files()
    import audformat
    df_test = audformat.utils.read_csv('speechrate_tests.csv')
    files = df_test.index.values
    cols = ['file', 'nsyll', 'npause', 'dur(s)', 'phonationtime(s)', 'speechrate(nsyll / dur)', 'articulation '
            'rate(nsyll / phonationtime)', 'ASD(speakingtime / nsyll)']
    datalist = []
    for index, file in enumerate(files):
        # print(f'processing {file}')
        speechrate_dictionary = speech_rate(file)
        datalist.append(speechrate_dictionary)
        if index%10==0:
            print('.', end=' ')
    df = pd.DataFrame(datalist)
    print(f'processed {index} files')
    df.to_csv('speechrate_data.csv', index=0)


import nkululeko.glob_conf as glob_conf
from nkululeko.util import Util
import numpy as np
import sounddevice as sd
import audiofile

class Demo_predictor():


    def __init__(self, model, file, is_list, feature_extractor, label_encoder):
        """Constructor setting up name and configuration"""
        self.model = model
        self.feature_extractor = feature_extractor
        self.label_encoder = label_encoder
        self.is_list = is_list
        self.sr = 16000
        self.target = glob_conf.config['DATA']['target']
        self.util = Util('demo_predictor')
        self.file = file

    def run_demo(self):
        signal, sr = None, 0
        if self.file is not None:
            if not self.is_list:
                sig, sr = audiofile.read(self.file)
                print(f'predicting file: {self.file}, len: {len(sig)} bytes, sampling rate: {sr}')
                self.predict_signal(sig, sr)
            else:
                with open(self.file) as f:
                    for index, line in enumerate(f):
                        sig, sr = audiofile.read(line.strip())
                        print(f'predicting file {index}: {line.strip()}')
                        self.predict_signal(sig, sr)                    
        else:
            while True:
                signal = self.record_audio(3)
                self.predict_signal(signal, self.sr)
    #            self.play_audio(signal)


    def predict_signal(self, signal, sr):
        features = self.feature_extractor.extract_sample(signal, sr)
        result_dict = self.model.predict_sample(features)
        keys = result_dict.keys()
        if self.label_encoder is not None:
            dict_2 = {}
            for i, k in enumerate(keys):
                ak = np.array(int(k)).reshape(1)
                lab = self.label_encoder.inverse_transform(ak)[0]
                dict_2[lab] = f'{result_dict[k]:.3f}'
            print(dict_2)
        else:
            print(result_dict)

    def record_audio(self, seconds): 
        print("recording ...") 
        y = sd.rec(int(seconds * self.sr), samplerate=self.sr, channels=1) 
        sd.wait() 
        y = y.T 
        return y 

    def play_audio(self, signal): 
        print("playback ...") 
        sd.play(signal.T, self.sr)
        status = sd.wait()


# global_config.py

def init_config(config_obj):
    global config
    config = config_obj

def set_label_encoder(encoder):
    global label_encoder
    label_encoder = encoder

def set_util(util_obj):
    global util
    util = util_obj